/***************************************************************************
 *
 * Copyright (c) 2003 Promise Technology, Inc.  All rights reserved.
 * No part of this document may be reproduced or transmitted in any form or
 * by any means, electronic or mechanical, for any purpose, without the
 * express written permission of Promise Technology, Inc.
 *
 *
 * cam_ei.c - functions for Error injection, simulation
 *
 ***************************************************************************/
#ifdef _ERR_INJECT_

#include "cam_con.h"
#include "cam_def.h"
#include "cam_imp.h"
#include "cam_macr.h"
#include "cam_gb.h"
#include "cam_exp.h"


/*=============================================================*/
/* Allocatie a free Err Injected Task, return a free Err Inkected Task ID */
/*=============================================================*/
U16 EI_Alloc_ErrInj_Task(void)
{
	U16 	i;
	
	for(i=0; i<MAX_ERRINJ_TASK; i++){	
		if(ErrInj_Buffer[i].UsedFlag == FALSE){
			ErrInj_Buffer[i].UsedFlag = TRUE;
			break;
		}	
	}
	
	if(i >= MAX_ERRINJ_TASK)
		return U16NULL;
	else
		return i;
}


/*==================================*/
/* Free Err Injected Task, no return value */
/*==================================*/
void EI_Free_ErrInj_Task(U16 Task_Id)
{
	if(ErrInj_Buffer[Task_Id].UsedFlag == TRUE)
		ErrInj_Buffer[Task_Id].UsedFlag = FALSE;

	ErrInj_Buffer[Task_Id].NextTaskId = U16NULL;	

	return;
}

/*=============================================================================*/
/* Insert a Err Injected Task to channel Q, and sorted by LBA, then mark the Task to Used */
/*=============================================================================*/
void EI_Insert2ChQ(U8 bCh, U16 Task_Id)
{
	U32	Inj_LBA;
	U16	tmp_current, tmp_next;
	
	if(ErrInj_Buffer_Table[bCh].head == U16NULL){
		ErrInj_Buffer_Table[bCh].head = Task_Id;	
		ErrInj_Buffer_Table[bCh].tail = Task_Id;
		ErrInj_Buffer[Task_Id].NextTaskId = U16NULL;		
	}
	else{
		tmp_next = ErrInj_Buffer_Table[bCh].head;
		Inj_LBA = ErrInj_Buffer[Task_Id].ErrInjTask_Config.HitLBAl;
		
		while(tmp_next != U16NULL){
			if(Inj_LBA < ErrInj_Buffer[tmp_next].ErrInjTask_Config.HitLBAl){				
				if(tmp_next == ErrInj_Buffer_Table[bCh].head){
					ErrInj_Buffer_Table[bCh].head = Task_Id;
					ErrInj_Buffer[Task_Id].NextTaskId = tmp_next;
				}
				else{
					ErrInj_Buffer[tmp_current].NextTaskId = Task_Id;
					ErrInj_Buffer[Task_Id].NextTaskId = tmp_next;				
				}
				break;
			}
			tmp_current = tmp_next;
			tmp_next = ErrInj_Buffer[tmp_current].NextTaskId;
		}
		if(tmp_next == U16NULL){
			ErrInj_Buffer[ErrInj_Buffer_Table[bCh].tail].NextTaskId = Task_Id;
			ErrInj_Buffer_Table[bCh].tail = Task_Id;
			ErrInj_Buffer[Task_Id].NextTaskId = U16NULL;	
		}
	}
	ErrInj_Buffer_Table[bCh].count++;	

	if(ErrInj_Buffer[Task_Id].ErrInjTask_Config.TaskCount)
		ErrInj_Buffer_Table[bCh].status |= EI_NEED_COUNT_DOWN;
	
	return;	
}



/*======================================================*/
/* delete a Err Injected Task from channel Q, and free this Task */
/*======================================================*/
U8 EI_Delete2ChQ(U8 bCh, U16 Task_Id)
{
	U16	tmp_current, tmp_next;
	U8	ret_value=FALSE;
	
	if(Task_Id == U16NULL){ /* delete all element */
		tmp_current = ErrInj_Buffer_Table[bCh].head;
		while(tmp_current != U16NULL){			
			tmp_next = ErrInj_Buffer[tmp_current].NextTaskId;
			EI_Free_ErrInj_Task(tmp_current);
			tmp_current = tmp_next;
		}
		ErrInj_Buffer_Table[bCh].count = 0;
		ErrInj_Buffer_Table[bCh].head = U16NULL;
		ErrInj_Buffer_Table[bCh].tail = U16NULL;
		return TRUE;
	}


	if(ErrInj_Buffer_Table[bCh].count == 0)
		return ret_value;
		
	tmp_next = ErrInj_Buffer_Table[bCh].head;
	while(tmp_next != U16NULL){
		if(tmp_next == Task_Id){				
			if(tmp_next == ErrInj_Buffer_Table[bCh].head){
				ErrInj_Buffer_Table[bCh].head = ErrInj_Buffer[Task_Id].NextTaskId;
				if(tmp_next == ErrInj_Buffer_Table[bCh].tail)
					ErrInj_Buffer_Table[bCh].tail = ErrInj_Buffer_Table[bCh].head;
			}
			else{ 
				ErrInj_Buffer[tmp_current].NextTaskId = ErrInj_Buffer[tmp_next].NextTaskId;
				if(tmp_next == ErrInj_Buffer_Table[bCh].tail)
					ErrInj_Buffer_Table[bCh].tail = tmp_current;
			}
		
			ErrInj_Buffer_Table[bCh].count--;
			ret_value = TRUE;
				
			break;
		}	
		tmp_current = tmp_next;
		tmp_next = ErrInj_Buffer[tmp_current].NextTaskId;
	}
	return ret_value;	
}


/*=============================================*/
/* Disable PHY to generate a unplug ISR			   */
/*=============================================*/
void EI_Disable_PHY(U8 bChannelNumber)
{
	U8		j;
	U32		ldata;
	PTR_SIZE	BaseIOAddress = ATABaseAddr[bChannelNumber];

		
	if(ErrInject_UnPlug_flag[bChannelNumber] == FALSE){
		ErrInject_UnPlug_flag[bChannelNumber] = TRUE;

#ifdef	_MMIO_
		for (j=0;j<10;j++) {
			WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, 4 );
			camUStallExecution(1);
			ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 4)
				break;
		}
		camUStallExecution(1);
#else
		for (j=0;j<10;j++) {
			WriteIndexUlong(BaseIOAddress, iNDEXIDESATASControl, 4 );
			camUStallExecution(1);
			ReadIndexUlong(BaseIOAddress, iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 4)
				break;
		}
		camUStallExecution(1);
#endif		
	}
	return;
}


/*=============================================*/
/* Enable PHY to generate a plug ISR				   */
/*=============================================*/
void EI_Enable_PHY(U8 bChannelNumber)
{
	U8		j;
	U32		ldata;
	PTR_SIZE	BaseIOAddress = ATABaseAddr[bChannelNumber];

		
	if(ErrInject_UnPlug_flag[bChannelNumber] == TRUE){
		ErrInject_UnPlug_flag[bChannelNumber] = FALSE;

#ifdef	_MMIO_
		for (j=0;j<10;j++) {
			WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, 1 );
			camUStallExecution(1);
			ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 1)
				break;
		}
		camUStallExecution(100);
		for (j=0;j<10;j++) {
			WriteIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, 0 );
			camUStallExecution(1);
			ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL] + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 0)
				break;			
		}
		camUStallExecution(100);
#else
		for (j=0;j<10;j++) {
			WriteIndexUlong(BaseIOAddress, iNDEXIDESATASControl, 1 );
			camUStallExecution(1);
			ReadIndexUlong(BaseIOAddress, iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 1)
				break;
		}
		camUStallExecution(100);
		for (j=0;j<10;j++) {
			WriteIndexUlong(BaseIOAddress, iNDEXIDESATASControl, 0 );
			camUStallExecution(1);
			ReadIndexUlong(BaseIOAddress, iNDEXIDESATASControl, ldata );
			if ( (ldata&0x07) == 0)
				break;			
		}
		camUStallExecution(100);
#endif
	}
	return;
}


/*=============================================*/
/* export to channel driver to injected a Error in CAM */
/*=============================================*/
U16 CAM_ErrorInjection(PERR_CONFIG pErrConfig)
{
	U8		bChannelNumber=CAM_ATA_Chnl_InMAP(pErrConfig->bCh);
	U8		bID=pErrConfig->bID;
	U16		i;
	U16		ErrInjTask_Id;
	
	
	if ( gChnlConfig[bChannelNumber].bCh == camCHNL_NOT_FOUND ){
		return 0xFF;	
	}


	if (!(gDrvConfig[bChannelNumber][bID].DevFlag & DEV_EXIST) ) {
		if(ErrInject_UnPlug_flag[bChannelNumber] == TRUE){
			if(pErrConfig->ErrType == ERRINJ_SATA_PLUG){
				if(gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA)
					return 0xFF;
					
				EI_Enable_PHY(bChannelNumber);
				return (U16)MAX_ERRINJ_TASK+1;				
			}	
		}
		return 0xFF;		
	}
	if (!(gDrvConfig[bChannelNumber][bID].DevFlag & DEV_ATA_DEVICE) ) {
		return 0xFF;		
	}

	if(!pErrConfig->ErrType)
		return 0xFF;		

	if(pErrConfig->ErrType == ERRINJ_SATA_UNPLUG){
		if(gChnlConfig[bChannelNumber].ChnlType != CHNL_SATA)
			return 0xFF;
			
		EI_Disable_PHY(bChannelNumber);
		return (U16)MAX_ERRINJ_TASK+1;
	}
	
	/* first call this function, initial the task buffer  */
	if(First_Enter_ErrInj == TRUE){
		First_Enter_ErrInj = FALSE;
		for(i=0; i<MAX_ERRINJ_TASK; i++)
			ErrInj_Buffer[i].UsedFlag = FALSE;
	}
		
		
	/* get free ERRInj_TASK */
	ErrInjTask_Id = EI_Alloc_ErrInj_Task();
	if(ErrInjTask_Id == U16NULL)
		return 0xFF;
	
	/* copy User injected data */
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.bCh = bChannelNumber;
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.bID = bID;
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.TaskCount = pErrConfig->TaskCount;
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.Flag = pErrConfig->Flag;
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.ErrType = pErrConfig->ErrType;
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.HitLBAl = pErrConfig->HitLBAl;
	ErrInj_Buffer[ErrInjTask_Id].ErrInjTask_Config.HitLBAh = pErrConfig->HitLBAh;

	/* check conflic for this Task -- type, lba,... */ 
	
	/* Insert to ErrInj Q */
	EI_Insert2ChQ(bChannelNumber, ErrInjTask_Id);		
	
	return ErrInjTask_Id;
}


/*==================================================*/
/* export to channel driver to clear a injected Error in CAM, */ 
/* if ErrHandle = U16NULL will delete all Task in channel Q  */
/*==================================================*/
U8 CAM_ClearErrorInjection(U8 Channel, U16 ErrHandle)
{
	Channel=CAM_ATA_Chnl_InMAP(Channel);

	if(ErrHandle == U16NULL)
		EI_Delete2ChQ(Channel, ErrHandle);
	else	
		if((EI_Delete2ChQ(Channel, ErrHandle)) == TRUE)
			EI_Free_ErrInj_Task(ErrHandle);
		
	return camACCEPTED;
}




void CheckErrInjection(PCAM_PACKET CAM_ATAPacket, PATA_TASK pATATask)
{
	U16	tmp_current, tmp_next, tmp_tail;
	U8	Hit_flag=FALSE;
	U8	bCh=pATATask->bCh2Chip;
	U8	LBA_PreCheck=FALSE;
	U8	NeedCountDown=FALSE;
	
	if (!(pATATask->ATACmdFlag & camDATA_XFER)) 
		return;
	if (!(pATATask->ATACmdFlag & ATA_DEVICE))	
		return;
		
	tmp_next = ErrInj_Buffer_Table[bCh].head;

	if(!(ErrInj_Buffer_Table[bCh].status & EI_NEED_COUNT_DOWN)){
		while(tmp_next != U16NULL){
			tmp_current = tmp_next;
			tmp_next = ErrInj_Buffer[tmp_current].NextTaskId;
			
			/*if(pATATask->bCh == ErrInj_Buffer[tmp_current].ErrInjTask_Config.bCh){*/
			if(pATATask->bID == ErrInj_Buffer[tmp_current].ErrInjTask_Config.bID){
				switch(ErrInj_Buffer[tmp_current].ErrInjTask_Config.ErrType){
					case ERRINJ_PH_ERR:
					case ERRINJ_SH_ERR:
					case ERRINJ_DH_ERR:
					case ERRINJ_HTO_ERR:
					case ERRINJ_TA_HBA_ERR:
					case ERRINJ_TA_DMA_CNT_ERR:
					case ERRINJ_PCI:
					case ERRINJ_ATA_IDNF:					
					case ERRINJ_ATA_DF:
						CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;
						Hit_flag = TRUE;
						pATATask->Reserved3 = tmp_current;
						break;					
					
					case ERRINJ_ATA_CRC:
					case ERRINJ_OVERRUN:
					case ERRINJ_UNDERRUN:
						if(!(pATATask->ATACmdFlag & PIO_XFER)){
							CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;
							Hit_flag = TRUE;
							pATATask->Reserved3 = tmp_current;
						}
						break;					

					case ERRINJ_TIMEOUT:
						CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;
						CAM_ATAPacket->PacketHead.Flag |= PACFLAG_TIMEOUT_INJECT;						
						CAM_ATAPacket->ATAPacket.bATRL[6] |= ATRLREG_X_END_1FX;
						Hit_flag = TRUE;
						if((EI_Delete2ChQ(bCh, tmp_current)) == TRUE)
							EI_Free_ErrInj_Task(tmp_current);
						break;					
					
					case ERRINJ_ATA_UNC:
					case ERRINJ_ATA_UNC_W:											
						if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.Flag & ERR_AUTO_HIT_LBA){							
							/* LBA auto hit case*/
							EI_Delete2ChQ(bCh, tmp_current);						
							
							ErrInj_Buffer[tmp_current].ErrInjTask_Config.Flag &= ~ERR_AUTO_HIT_LBA;
							ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl = pATATask->Cmd.ATACmd.LBAl;

							EI_Insert2ChQ(bCh, tmp_current);						
						}
						
						if(LBA_PreCheck == FALSE){
							LBA_PreCheck = TRUE;
							tmp_tail = ErrInj_Buffer_Table[bCh].tail;							
							if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl > (pATATask->Cmd.ATACmd.LBAl+pATATask->Cmd.ATACmd.bCount)){
								Hit_flag = TRUE;
								break;
							}
							if(ErrInj_Buffer[tmp_tail].ErrInjTask_Config.HitLBAl < (pATATask->Cmd.ATACmd.LBAl)){
								if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl != 0){
									Hit_flag = TRUE;					
									break;
								}
							}
						}
						
						if((ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl >= pATATask->Cmd.ATACmd.LBAl) &&
						 (ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl <= (pATATask->Cmd.ATACmd.LBAl+pATATask->Cmd.ATACmd.bCount))){
							if(pATATask->ATACmdFlag & camDATA_IN){/* read cmd */
								CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;							
								pATATask->Reserved3 = tmp_current;
								Hit_flag = TRUE;
							}						
							else{/* write cmd */
								if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.ErrType == ERRINJ_ATA_UNC){
									if((EI_Delete2ChQ(bCh, tmp_current)) == TRUE)
										EI_Free_ErrInj_Task(tmp_current);
								}
								else{
									CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;							
									pATATask->Reserved3 = tmp_current;
									Hit_flag = TRUE;
								}
							}
						}
						break;
												
					default:
						break;
				}/* end switch() */
				if(Hit_flag == TRUE)
					break;
			}/* end of if() */			
		}/* end of while() */
	}/* end of if() */
	else{ /* (ErrInj_Buffer_Table[bCh].status & EI_NEED_COUNT_DOWN) */
		while(tmp_next != U16NULL){
			tmp_current = tmp_next;
			tmp_next = ErrInj_Buffer[tmp_current].NextTaskId;
			
			/*if(pATATask->bCh == ErrInj_Buffer[tmp_current].ErrInjTask_Config.bCh){*/
			if(pATATask->bID == ErrInj_Buffer[tmp_current].ErrInjTask_Config.bID){
				if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.TaskCount){
					ErrInj_Buffer[tmp_current].ErrInjTask_Config.TaskCount--;
					if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.TaskCount)
						NeedCountDown = TRUE;	
				}
				else if(Hit_flag != TRUE){
					switch(ErrInj_Buffer[tmp_current].ErrInjTask_Config.ErrType){
						case ERRINJ_PH_ERR:
						case ERRINJ_SH_ERR:
						case ERRINJ_DH_ERR:
						case ERRINJ_HTO_ERR:
						case ERRINJ_TA_HBA_ERR:
						case ERRINJ_TA_DMA_CNT_ERR:
						case ERRINJ_PCI:
						case ERRINJ_ATA_IDNF:					
						case ERRINJ_ATA_DF:
							CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;
							Hit_flag = TRUE;
							pATATask->Reserved3 = tmp_current;
							break;					
						
						case ERRINJ_ATA_CRC:
						case ERRINJ_OVERRUN:
						case ERRINJ_UNDERRUN:
							if(!(pATATask->ATACmdFlag & PIO_XFER)){
								CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;
								pATATask->Reserved3 = tmp_current;
								Hit_flag = TRUE;
							}
							break;					
	
						case ERRINJ_TIMEOUT:
							CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;
							CAM_ATAPacket->PacketHead.Flag |= PACFLAG_TIMEOUT_INJECT;
							CAM_ATAPacket->ATAPacket.bATRL[6] |= ATRLREG_X_END_1FX;
							Hit_flag = TRUE;
							if((EI_Delete2ChQ(bCh, tmp_current)) == TRUE)
									EI_Free_ErrInj_Task(tmp_current);
							break;					
						
						case ERRINJ_ATA_UNC:
						case ERRINJ_ATA_UNC_W:											
							if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.Flag & ERR_AUTO_HIT_LBA){							
								/* LBA auto hit case*/
								EI_Delete2ChQ(bCh, tmp_current);						
								
								ErrInj_Buffer[tmp_current].ErrInjTask_Config.Flag &= ~ERR_AUTO_HIT_LBA;
								ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl = pATATask->Cmd.ATACmd.LBAl;
	
								EI_Insert2ChQ(bCh, tmp_current);						
							}		

							if(LBA_PreCheck == FALSE){
								LBA_PreCheck = TRUE;
								tmp_tail = ErrInj_Buffer_Table[bCh].tail;							
								if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl > (pATATask->Cmd.ATACmd.LBAl+pATATask->Cmd.ATACmd.bCount)){
									Hit_flag = TRUE;
									break;
								}
								if(ErrInj_Buffer[tmp_tail].ErrInjTask_Config.HitLBAl < (pATATask->Cmd.ATACmd.LBAl)){
									if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl != 0){
										Hit_flag = TRUE;					
										break;
									}
								}
							}
							
							if((ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl >= pATATask->Cmd.ATACmd.LBAl) &&
							 (ErrInj_Buffer[tmp_current].ErrInjTask_Config.HitLBAl <= (pATATask->Cmd.ATACmd.LBAl+pATATask->Cmd.ATACmd.bCount))){
								if(pATATask->ATACmdFlag & camDATA_IN){/* read cmd */
									CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;							
									Hit_flag = TRUE;
									pATATask->Reserved3 = tmp_current;
								}						
								else{/* write cmd */
									if(ErrInj_Buffer[tmp_current].ErrInjTask_Config.ErrType == ERRINJ_ATA_UNC){
										if((EI_Delete2ChQ(bCh, tmp_current)) == TRUE)
											EI_Free_ErrInj_Task(tmp_current);
									}
									else{
										CAM_ATAPacket->PacketHead.Flag |= PACFLAG_ERR_INJECT;							
										Hit_flag = TRUE;
										pATATask->Reserved3 = tmp_current;
									}
								}
							}
							break;
													
						default:
							break;
					}/* end switch() */
				}/* end of else if() */
			}/* end of if() */			
		}/* end of while() */
		if(NeedCountDown != TRUE)
			ErrInj_Buffer_Table[bCh].status &= ~EI_NEED_COUNT_DOWN;
	}
	return;		
}






void ERR_CAM_NCQ_ISR(U8 Adapter_ID,U8 SEQID)
{
	PSEQ_INFO		pCurSEQ_INFO;
	U8 				ModuleType;
	U8				i;
	PCAM_PACKET ultmpATAPacketAddress, ultmpATAPacketAddress_next;
	PATA_TASK 		pATATask;
	U8 				SEQerrorFlag=FALSE;
	U8 				SEQCntCtrlErr;
	U8 				bChannelNumber;
	U32 			tmpData;
	U32 			SActiveData;
	U8 				TAG;
	U8			CRC_err=FALSE;
	U16			ErrInj_TaskId;

	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
	ModuleType= (U8)pCurSEQ_INFO->Owner;
	bChannelNumber = Adapter_ID*MAX_ATA_CHNL+ModuleType;
	
	SEQerrorFlag = FALSE;
						
	/* check error from Sequence Counter Control Register */
	ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
	SEQCntCtrlErr &= (1<<ModuleType);

/*=============================================================================================*/
	/* _ERR_INJECT_ */
	/* get finish TAGs */
	ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_SActive, SActiveData);
	SActiveData = (gChnlDrvControlConfig[bChannelNumber].UsedTAG & (~SActiveData));
	for (i=0;i<MAX_TAG;i++) {
		if (!((SActiveData >> i) & 0x01))
			continue;
		TAG=i;
		ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
		while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){
			if (((PCAM_PACKET)ultmpATAPacketAddress)->ATAPacket.bDelaySeqID==TAG){	
				if((ultmpATAPacketAddress->PacketHead.Flag & PACFLAG_ERR_INJECT)){
					/* hit error injection, clear mask */
					ErrInject_SEQ[Adapter_ID] &= ~((U16)(1 << SEQID));
					pATATask = (PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;					
					ErrInj_TaskId = pATATask->Reserved3;
					if(ErrInj_Buffer[ErrInj_TaskId].UsedFlag == TRUE)
						SEQCntCtrlErr = 1;

				}
				break;
			}
			ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
			
			ultmpATAPacketAddress = ultmpATAPacketAddress_next;		
		}
	}

/*=============================================================================================*/
				
	if (SEQCntCtrlErr) {
		SEQerrorFlag = TRUE;
	}
	else { /* non error */
		if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
				#ifdef	_MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
				#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
				#endif
				if (tmpData & CRC_ERR) {	/* CRC error */
					SEQerrorFlag = TRUE;
					CRC_err=TRUE;
				}
			}
		}
	}
									
	/* if no error(SATA error and packet error) and no data or ide register status need to transfer(PIO, ATA_STATUS_RETURN or ATAPI ), try to execute other packets.*/
	if (!(SEQerrorFlag == TRUE)) {
		if (pCurSEQ_INFO->GO==TRUE) {
			if (ATAModuleNeedPush[bChannelNumber]==TRUE){
				/* start ata module */
				StartATAOperation(bChannelNumber);
			}
		}
	}
	
	if (CRC_err==TRUE)
		SActiveData = 0;	/* no commands complete */
	else {
		/* get finish TAGs */
		ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_SActive, SActiveData);
		SActiveData = (gChnlDrvControlConfig[bChannelNumber].UsedTAG & (~SActiveData));
	}
			
	/* when NCQ running, can't write FPDMA Control/Status Register whole DWORD */
	/* only clear the FPDMA Control/Status Register bit 10, 11*/
	ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status, tmpData);
	tmpData |= FPDMA_Ctrl_Status_DMAsetup_INT_flag | FPDMA_Ctrl_Status_SetDb_INT_flag;
	WriteIndexUchar(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_FPDMA_Ctrl_Status+1, (U8) (tmpData>> 8));

	if (SEQerrorFlag == TRUE) { /* error */
	
		camStallExecution(1);
		
		/* get ATA Global control & Status Register */
		ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);

		/* _ERR_INJECT_ */
		switch(ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType){
			case ERRINJ_PH_ERR:
				tmpData = ((U32)ATA_PH_ERR << 8);
				break;
			case ERRINJ_SH_ERR:
				tmpData = ((U32)ATA_SH_ERR << 8);	
				break;
		 	case ERRINJ_DH_ERR:
		 		tmpData = ((U32)ATA_DH_ERR << 8);
				break;
		 	case ERRINJ_HTO_ERR:	
		 		tmpData = ((U32)ATA_HTO_ERR << 8);
				break;
		 	case ERRINJ_TA_HBA_ERR:	
		 		tmpData = ((U32)ATA_TA_HBA_ERR << 8);
				break;
		 	case ERRINJ_TA_DMA_CNT_ERR:
		 		tmpData = ((U32)ATA_TA_DMA_CNT_ERR << 8);
				break;
		 	case ERRINJ_PCI:	
		 		tmpData = ((U32)ATAERR_PCI << 8);
				break;
		 	case ERRINJ_TIMEOUT:	
				break;
			case ERRINJ_OVERRUN:	
				tmpData = ((U32)ATAERR_OVERRUN << 8);
				break;
			case ERRINJ_UNDERRUN: 
				tmpData = ((U32)ATAERR_UNDERRUN << 8);
				break;
												
			case ERRINJ_ATA_IDNF:
			case ERRINJ_ATA_UNC:
			case ERRINJ_ATA_UNC_W:
			case ERRINJ_ATA_CRC:
			case ERRINJ_ATA_DF:
				tmpData = ((U32)ATAERR_DRV << 8);
				break;
			default:	
				break;
		}			
		
		/* soft reset ATA Module after getting status. */
		ATASoftResetModule(bChannelNumber);
		NeedCheckFIFO[bChannelNumber] = TRUE;
		if((ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType != ERRINJ_ATA_UNC)
		&& (ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType != ERRINJ_ATA_UNC_W)){
			if((EI_Delete2ChQ(bChannelNumber, ErrInj_TaskId)) == TRUE)
				EI_Free_ErrInj_Task(ErrInj_TaskId);
		}

	}
	else{				
		for (i=0;i<MAX_TAG;i++) {
			if (!((SActiveData >> i) & 0x01))
				continue;
			TAG=i;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
			pATATask->ATAStatus= 0;		/* default no error */	
			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);				
			pATATask->callback(pATATask);	
		}
	}
	
	if (SEQerrorFlag == TRUE) { /* error */
	
		PCAM_PACKET ultmpATAPacketAddressHead;

		
		
		
		TAG=U8NULL;
		
		/* if can't get the error TAG, hard reset and abort all tasks */
		if (TAG==U8NULL)
		{
			U8	receiveStatus;
			
			/* disable receive task.*/
			receiveStatus=ATAModuleReceiveTask[bChannelNumber];
			ATAModuleReceiveTask[bChannelNumber]=FALSE;
				
			ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;			
			/* return all execute commands with error */
			Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead,(U16)((tmpData >> 8)|ATA_ERROR) );
			ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;
			ATAModuleBusy[bChannelNumber]=FALSE;	
			gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
			
			/* do hard reset so ASIC can restart again */
			/*CAM_Reset_ReinitATA(bChannelNumber);*/
			ATASoftResetModule(bChannelNumber);
			ATAModuleReceiveTask[bChannelNumber]=receiveStatus;
						
		}
		else
		{
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
			if (ultmpATAPacketAddress != (PCAM_PACKET)PTR_NULL) {
				
				ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
				pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
				pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
				/* back undo execute packets to wait queue. */
				Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead);
				ATAModuleBusy[bChannelNumber]=FALSE;
				gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
				gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
				ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;
								
				pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
	
				pATATask->ATAStatus = (U16)(tmpData >> 8);
				pATATask->ATAStatus &= ATAERR_HWERROR;
				pATATask->ATAStatus |= ATA_ERROR;						
	
				GetStatus(ATABaseAddr[pATATask->bCh2Chip], pATATask->Cmd.ATACmd.bStatus);
				
				if (pATATask->Cmd.ATACmd.bStatus & IDE_STATUS_BUSY){
					pATATask->ATAStatus &= ~ATAERR_DRV;
					pATATask->ATAStatus |= ATA_TIMEOUT;
				}
				else{				
					/* set disk status */
					if (pATATask->ATACmdFlag & ATA_DEVICE)									
						ATAErrorStatus(pATATask);
				}
				
				/* _ERR_INJECT_ */
				if (pATATask->ATACmdFlag & ATA_DEVICE){
					pATATask->Cmd.ATACmd.bStatus = 0x0; 
					pATATask->Cmd.ATACmd.bError = 0x0;
	
					switch(ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType){
						case ERRINJ_ATA_IDNF:
							pATATask->Cmd.ATACmd.bStatus = 0x51; 
							pATATask->Cmd.ATACmd.bError = 0x14;
							break;
						case ERRINJ_ATA_UNC:
							pATATask->Cmd.ATACmd.bStatus = 0x51; 
							pATATask->Cmd.ATACmd.bError = 0x44;
							pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
							break;	
						case ERRINJ_ATA_UNC_W:
							if(pATATask->ATACmdFlag & camDATA_IN){
								pATATask->Cmd.ATACmd.bStatus = 0x51; 
								pATATask->Cmd.ATACmd.bError = 0x44;
								pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
							}
							else{
								pATATask->Cmd.ATACmd.bStatus = 0x70; 
								pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
							}
							break;
						case ERRINJ_ATA_CRC:
							pATATask->Cmd.ATACmd.bStatus = 0x51; 
							pATATask->Cmd.ATACmd.bError = 0x84;
							break;
						case ERRINJ_ATA_DF:
							pATATask->Cmd.ATACmd.bStatus = 0x70; 
							break;
						default:	
							break;
					}			
				}


				/* free sg resource */
				CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
				/* free packet resource. */
				CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
				pATATask->callback(pATATask);	
			}
			else {
				ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
				pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
				pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;				
				/* return all execute commands with error */
				Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead,(U16)((tmpData >> 8)|ATA_ERROR));
				ATAModuleBusy[bChannelNumber]=FALSE;
				ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;	
				gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
				gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;								
			}
		}	

	}

	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if (ATAModuleNeedPush[i]==TRUE){
			/* start ata module */
			StartATAOperation(i);
				break;
		}
	}
	return;
}



void ERR_CAM_QUEUED_ISR(U8 Adapter_ID,U8 SEQID)
{
	PSEQ_INFO		pCurSEQ_INFO;
	U8 				ModuleType;
	U8				i;
	PCAM_PACKET ultmpATAPacketAddress, ultmpATAPacketAddress_next;
	PATA_TASK 		pATATask;
	U8 				SEQCntCtrlErr;
	U8 				bChannelNumber;
	U8 				TAG;
	U8				Status;
	U32				tmpData;
	U8				CRC_err=FALSE, ServiceFlag;
	U16				ErrInj_TaskId;

	U32				StatusPollingCount;
	U8				ErrorOccurFlag;			


	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
	ModuleType= (U8)pCurSEQ_INFO->Owner;
	bChannelNumber = Adapter_ID*MAX_ATA_CHNL+ModuleType;
	
	/* check error from Sequence Counter Control Register */
	ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
	SEQCntCtrlErr &= (1<<ModuleType);


/*=============================================================================================*/
	/* _ERR_INJECT_ */
	/* get the TAG */
	TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
	TAG = (TAG >> 3);

	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
	while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){
		if (((PCAM_PACKET)ultmpATAPacketAddress)->ATAPacket.bDelaySeqID==TAG){	
			if((ultmpATAPacketAddress->PacketHead.Flag & PACFLAG_ERR_INJECT)){
					/* hit error injection, clear mask */
				ErrInject_SEQ[Adapter_ID] &= ~((U16)(1 << SEQID));
				pATATask = (PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;					
				ErrInj_TaskId = pATATask->Reserved3;
				if(ErrInj_Buffer[ErrInj_TaskId].UsedFlag == TRUE)
					SEQCntCtrlErr = 1;
			}
			break;
		}
		ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
		
		ultmpATAPacketAddress = ultmpATAPacketAddress_next;		
	}

/*=============================================================================================*/



	if (!SEQCntCtrlErr) { /* non error */
		if(gAdapterConfig[Adapter_ID].TapeOutVersion < REVISION_2){/* ASIC tapout prior revision 2 */
			if (gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA) {/* SATA */
				#ifdef	_MMIO_
					ReadIndexUlong(ATABaseAddr[(bChannelNumber/MAX_ATA_CHNL)*MAX_ATA_CHNL]  + 0x100 * (bChannelNumber%MAX_ATA_CHNL) , iNDEXIDESATAError, tmpData );
				#else
					ReadIndexUlong(ATABaseAddr[bChannelNumber], iNDEXIDESATAError, tmpData );
				#endif
				if (tmpData & CRC_ERR) {	/* CRC error */
					CRC_err=TRUE;
					SEQCntCtrlErr=TRUE;
				}
			}
		}
	}
					
	if (!SEQCntCtrlErr) {

		/* read status */
		Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
		if (Status & IDE_STATUS_BUSY)
			return;

		if (gChnlDrvControlConfig[bChannelNumber].TCQTransferData==TRUE) {				
			/* command complete */
			/* get the TAG */
			if(gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG == U8NULL){
				TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
				TAG = (TAG >> 3);
			}
			else
				TAG = gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG;
				
			gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
			gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG = U8NULL;
			
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
			if (ultmpATAPacketAddress == (PCAM_PACKET)PTR_NULL)
				return;
							
			gChnlDrvControlConfig[bChannelNumber].TCQTransferData = FALSE;
			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
			pATATask->ATAStatus= 0;		/* default no error */	

			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
			pATATask->callback(pATATask);	
		}

		if (gChnlDrvControlConfig[bChannelNumber].TCQTransferData != TRUE) {	
			ServiceFlag = FALSE;
			StatusPollingCount = 0;
			ErrorOccurFlag = FALSE;
			if (Status & IDE_STATUS_SERVICE) { /* Service release case */
				ServiceFlag = TRUE;
			}
			else{
				if(gChnlDrvControlConfig[bChannelNumber].IsWDHardDrive != TRUE){/* not Hitachi */
					if((ATAModuleBusy[bChannelNumber] == TRUE) && (ATAModuleNeedPush[bChannelNumber] != TRUE)){
						while(1) {
							if(StatusPollingCount >= 100000){	/* polling above 1 sec */
								ErrorOccurFlag = TRUE;
								break;
							}

							if (!(Status & IDE_STATUS_BUSY)) {
								if (Status & IDE_STATUS_SERVICE) {
									ServiceFlag = TRUE;
									break;
								}
							}
							camUStallExecution(1);
							Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
							StatusPollingCount++;
						}		
					}
				}
			}
	
			if(ServiceFlag == TRUE){
				gChnlDrvControlConfig[bChannelNumber].TCQTransferData=TRUE;
				/* send service command */
				IssueService(bChannelNumber,0,IDE_COMMAND_SERVICE);
	
				Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
				StatusPollingCount = 0;
				ErrorOccurFlag = FALSE;

				/* wait DRQ */	
				while(1) {
					if(StatusPollingCount >= 100000){	/* polling above 1 sec */
						ErrorOccurFlag = TRUE;
						break;
					}

					if (Status & IDE_STATUS_DRQ) {
						break;
					}
					camUStallExecution(1);
					Status = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDEStatus);
					StatusPollingCount++;
				}
									
				if(ErrorOccurFlag != TRUE){
					/* get the TAG */
					TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
					TAG = (TAG >> 3);
					gChnlDrvControlConfig[bChannelNumber].CurrentXferTAG = TAG;
					/* prepare SG */
		
					/* return the correspond packet from gSEQ_INFO */
					ultmpATAPacketAddress=ReturnATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
					if (ultmpATAPacketAddress == (PCAM_PACKET)PTR_NULL)
						return;
								
					pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
					
					/* write SG table */
					WriteIndexUlong(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEDestriptorPointer, ultmpATAPacketAddress->ATAPacket.ulPSG);
			
					/* start DMA transfer */
					if (pATATask->ATACmdFlag & camDATA_IN) /* read */
						WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|SEQID) );		
					else /* write */
						WriteIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECICR, (U8)(IDECICR_DMA_Start|IDECICR_DMA_RW_Direction|SEQID) );	
				}
			}
			
			if(ErrorOccurFlag == TRUE){
				if (ATA_Timeout_TimerID[bChannelNumber] != ZERO_NULL)
					/* remove ata timeout timer */
					camTimer(Adapter_ID,0,(void *)ATA_Timeout_TimerID[bChannelNumber]);	

				ATA_Timeout(bChannelNumber);
				return;
			}
		}
	}
	else { /* error */

		PCAM_PACKET ultmpATAPacketAddressHead;
		NeedCheckFIFO[bChannelNumber] = TRUE;
		
		/* soft reset ATA Module after getting status. */
		ATASoftResetModule(bChannelNumber);
		
		/* get the TAG */
		TAG = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount);
		TAG = (TAG >> 3);
		gChnlDrvControlConfig[bChannelNumber].UsedTAG &= ~(1 << TAG);
		gChnlDrvControlConfig[bChannelNumber].FreeTAG |= (1 << TAG);
		
		/* get ATA Global control & Status Register */
		ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);


		/* _ERR_INJECT_ */	
		switch(ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType){
			case ERRINJ_PH_ERR:
				tmpData = ((U32)ATA_PH_ERR << 8);
				break;
			case ERRINJ_SH_ERR:
				tmpData = ((U32)ATA_SH_ERR << 8);	
				break;
		 	case ERRINJ_DH_ERR:
		 		tmpData = ((U32)ATA_DH_ERR << 8);
				break;
		 	case ERRINJ_HTO_ERR:	
		 		tmpData = ((U32)ATA_HTO_ERR << 8);
				break;
		 	case ERRINJ_TA_HBA_ERR:	
		 		tmpData = ((U32)ATA_TA_HBA_ERR << 8);
				break;
		 	case ERRINJ_TA_DMA_CNT_ERR:
		 		tmpData = ((U32)ATA_TA_DMA_CNT_ERR << 8);
				break;
		 	case ERRINJ_PCI:	
		 		tmpData = ((U32)ATAERR_PCI << 8);
				break;
		 	case ERRINJ_TIMEOUT:	
				break;
			case ERRINJ_OVERRUN:	
				tmpData = ((U32)ATAERR_OVERRUN << 8);
				break;
			case ERRINJ_UNDERRUN: 
				tmpData = ((U32)ATAERR_UNDERRUN << 8);
				break;
												
			case ERRINJ_ATA_IDNF:
			case ERRINJ_ATA_UNC:
			case ERRINJ_ATA_UNC_W:
			case ERRINJ_ATA_CRC:
			case ERRINJ_ATA_DF:
				tmpData = ((U32)ATAERR_DRV << 8);
				break;
			default:	
				break;
		}			
		
		if((ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType != ERRINJ_ATA_UNC)
		&& (ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType != ERRINJ_ATA_UNC_W)){	
			if((EI_Delete2ChQ(bChannelNumber, ErrInj_TaskId)) == TRUE)
				EI_Free_ErrInj_Task(ErrInj_TaskId);
		}


		
		if (CRC_err)
			ultmpATAPacketAddress=(PCAM_PACKET)PTR_NULL;
		else							
			/* get PCAM_ATA_PACKET from SEQ_INFO */
			ultmpATAPacketAddress=GetATA_PacketFromSEQ_INFO(bChannelNumber,SEQID,TAG);
		
		gChnlDrvControlConfig[bChannelNumber].TCQTransferData=FALSE;
		
		if (ultmpATAPacketAddress != (PCAM_PACKET)PTR_NULL) {

			ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			ATAModuleBusy[bChannelNumber]=FALSE;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
			/* back undo execute packets to wait queue. */
			Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead);
			ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;
			gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;
											
			pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;

			pATATask->ATAStatus = (U16)(tmpData >> 8);
			pATATask->ATAStatus &= ATAERR_HWERROR;
			pATATask->ATAStatus |= ATA_ERROR;						

			GetStatus(ATABaseAddr[pATATask->bCh2Chip], pATATask->Cmd.ATACmd.bStatus);
			
			if (pATATask->Cmd.ATACmd.bStatus & IDE_STATUS_BUSY){
				pATATask->ATAStatus &= ~ATAERR_DRV;
				pATATask->ATAStatus |= ATA_TIMEOUT;
			}
			else{				
				/* set disk status */
				if (pATATask->ATACmdFlag & ATA_DEVICE)									
					ATAErrorStatus(pATATask);
			}		


			/* _ERR_INJECT_ */
			if (pATATask->ATACmdFlag & ATA_DEVICE){
				pATATask->Cmd.ATACmd.bStatus = 0x0; 
				pATATask->Cmd.ATACmd.bError = 0x0;

				switch(ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType){
					case ERRINJ_ATA_IDNF:
						pATATask->Cmd.ATACmd.bStatus = 0x51; 
						pATATask->Cmd.ATACmd.bError = 0x14;
						break;
					case ERRINJ_ATA_UNC:
						pATATask->Cmd.ATACmd.bStatus = 0x51; 
						pATATask->Cmd.ATACmd.bError = 0x44;
						pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
						break;	
					case ERRINJ_ATA_UNC_W:
						if(pATATask->ATACmdFlag & camDATA_IN){
							pATATask->Cmd.ATACmd.bStatus = 0x51; 
							pATATask->Cmd.ATACmd.bError = 0x44;
							pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
						}
						else{
							pATATask->Cmd.ATACmd.bStatus = 0x70; 
							pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
						}
						break;
					case ERRINJ_ATA_CRC:
						pATATask->Cmd.ATACmd.bStatus = 0x51; 
						pATATask->Cmd.ATACmd.bError = 0x84;
						break;
					case ERRINJ_ATA_DF:
						pATATask->Cmd.ATACmd.bStatus = 0x70; 
						break;
					default:	
						break;
				}			
			}

			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
			/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
			pATATask->callback(pATATask);	
		}
		else {
			ultmpATAPacketAddressHead = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
			pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
			pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;			
			/* return all execute commands with error */
			Return_Execute_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddressHead,(U16)((tmpData >> 8)|ATA_ERROR));
			ATAModuleBusy[bChannelNumber]=FALSE;
			ultmpATAPacketAddressHead=(PCAM_PACKET)PTR_NULL;	
			gChnlDrvControlConfig[bChannelNumber].FreeTAG = 0xFFFFFFFF;
			gChnlDrvControlConfig[bChannelNumber].UsedTAG = 0x0;						
		}
	}
	
	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if (ATAModuleNeedPush[i]==TRUE){
			/* start ata module */
			StartATAOperation(i);
				break;
		}
	}
	return;
}




void ERR_CAM_NORMAL_ISR(U8 Adapter_ID,U8 SEQID)
{
	PSEQ_INFO		pCurSEQ_INFO;
	U8 				ModuleType;
	U8				i;
	PCAM_PACKET ultmpATAPacketAddress, ultmpATAPacketAddress_next;
	PATA_TASK 		pATATask;
	U8 				SEQCntCtrlErr, SEQCntCtrl;
	U8 				bChannelNumber;
	U32 			tmpData;
	U8			bdata;
	U16		ErrInj_TaskId=0;

	/* Flag variable */
	U8 	RestartModule=FALSE;
	U8	FreePacketSource = TRUE;
	U8	errorFlag=FALSE;
	U8	SEQerrorFlag=FALSE;
	U8	IntReason=0;


	pCurSEQ_INFO = &(gSEQ_INFO[Adapter_ID][SEQID]);
	ModuleType= (U8)pCurSEQ_INFO->Owner;
	bChannelNumber = Adapter_ID*MAX_ATA_CHNL+ModuleType;
	
	SEQerrorFlag = FALSE;	
	RestartModule = FALSE;
	

	/* remove ata timeout timer */
	if (ATA_Timeout_TimerID[bChannelNumber] != ZERO_NULL) {
		camTimer(Adapter_ID,0,(void *)ATA_Timeout_TimerID[bChannelNumber]);	
		ATA_Timeout_TimerID[bChannelNumber]=ZERO_NULL;
	}
												
	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;

	/* read Sequence Counter Control Register # */
	ReadPortUchar(HostBaseAddr[Adapter_ID], SEQID*4, SEQCntCtrl);
	SEQCntCtrl &= 0x1F;

	/* _ERR_INJECT_ */
	for(i=1; i<=pCurSEQ_INFO->PACCNT; i++){
		if((ultmpATAPacketAddress->PacketHead.Flag & PACFLAG_ERR_INJECT)){
			pATATask = (PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;					
			ErrInj_TaskId = pATATask->Reserved3;
			if(ErrInj_Buffer[ErrInj_TaskId].UsedFlag == TRUE)
				break;
		}
		ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
		ultmpATAPacketAddress = ultmpATAPacketAddress_next;	
	}
	/* if(i > pCurSEQ_INFO->PACCNT)
		error */
	SEQCntCtrl = pCurSEQ_INFO->PACCNT - i;

	/* hit error injection, clear mask */
	ErrInject_SEQ[Adapter_ID] &= ~((U16)(1 << SEQID));

							
	/* check error from Sequence Counter Control Register */
	ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
	SEQCntCtrlErr &= (1<<ModuleType);
		
	if(ErrInj_Buffer[ErrInj_TaskId].UsedFlag == TRUE)
		SEQCntCtrlErr = 1;		

	if (SEQCntCtrlErr) {
		SEQerrorFlag = TRUE;
		/* ATAPI PIO SEQCntCtrl will be zero or if error, current packet maybe discount or not */
		if (SEQCntCtrl < pCurSEQ_INFO->PACCNT)
			SEQCntCtrl++;
	}

	ATAModuleSEQID[bChannelNumber] = U8NULL;
		
	ultmpATAPacketAddress = (PCAM_PACKET)pCurSEQ_INFO->SEQID2PacketHead;
	pCurSEQ_INFO->SEQID2PacketHead=(PVOID)PTR_NULL;
	pCurSEQ_INFO->SEQID2PacketTail=(PVOID)PTR_NULL;
								
	while ((PTR_SIZE)ultmpATAPacketAddress != PTR_NULL){						

		pATATask=(PATA_TASK)ultmpATAPacketAddress->PacketHead.PTA;
		pATATask->ATAStatus= 0;		/* default no error */
							
		/* initial control flag */
		FreePacketSource = TRUE;
		IntReason = 0;
		errorFlag = FALSE;
				
		if (SEQerrorFlag == TRUE)	
		{
			NeedCheckFIFO[bChannelNumber] = TRUE;
			/* check is error packet */ /* not success done all seq's packets  */
			if (pCurSEQ_INFO->PACCNT == SEQCntCtrl)
			{
				errorFlag = TRUE;								

				/* get ATA Global control & Status Register */
				ReadPortUlong(ATABaseAddr[bChannelNumber], ATA_STATUS_REGISTER_OFFSET, tmpData);
				/* soft reset ATA Module after getting status. */
				ATASoftResetModule(bChannelNumber);

				/* if swap box nop command, it will return error at all time, and it will not queue with others packets when execute,
									, needn't abort others packet at waiting queue */
				if (!(pATATask->ATACmdFlag & SWAPBOX_CMD)){
					if (pATATask->ATACmdFlag & ATA_DEVICE) { /* if ATAPI, needn't abort all packets */
						/* back undo packets to wait queue. */
						Back_All_Undone_ATAPackets(bChannelNumber,(PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA);
						ultmpATAPacketAddress->PacketHead.NPA=(PVOID)PTR_NULL;
					}
				}
				else /* SWAPBOX_CMD */
					goto SuccessCommandCompleted;
															

				/* _ERR_INJECT_ */	
				switch(ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType){
					case ERRINJ_PH_ERR:
						tmpData = ((U32)ATA_PH_ERR << 8);
						break;
					case ERRINJ_SH_ERR:
						tmpData = ((U32)ATA_SH_ERR << 8);	
						break;
				 	case ERRINJ_DH_ERR:
				 		tmpData = ((U32)ATA_DH_ERR << 8);
						break;
				 	case ERRINJ_HTO_ERR:	
				 		tmpData = ((U32)ATA_HTO_ERR << 8);
						break;
				 	case ERRINJ_TA_HBA_ERR:	
				 		tmpData = ((U32)ATA_TA_HBA_ERR << 8);
						break;
				 	case ERRINJ_TA_DMA_CNT_ERR:
				 		tmpData = ((U32)ATA_TA_DMA_CNT_ERR << 8);
						break;
				 	case ERRINJ_PCI:	
				 		tmpData = ((U32)ATAERR_PCI << 8);
						break;
				 	case ERRINJ_TIMEOUT:	
						break;
					case ERRINJ_OVERRUN:	
						tmpData = ((U32)ATAERR_OVERRUN << 8);
						break;
					case ERRINJ_UNDERRUN: 
						tmpData = ((U32)ATAERR_UNDERRUN << 8);
						break;
														
					case ERRINJ_ATA_IDNF:
					case ERRINJ_ATA_UNC:
					case ERRINJ_ATA_UNC_W:
					case ERRINJ_ATA_CRC:
					case ERRINJ_ATA_DF:
						tmpData = ((U32)ATAERR_DRV << 8);
						break;
					default:	
						break;
				}			
				
				if((ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType != ERRINJ_ATA_UNC)
				&& (ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType != ERRINJ_ATA_UNC_W)){	
					if((EI_Delete2ChQ(bChannelNumber, ErrInj_TaskId)) == TRUE)
						EI_Free_ErrInj_Task(ErrInj_TaskId);				
				}
	
				pATATask->ATAStatus = (U16)(tmpData >> 8);
															
	
				pATATask->ATAStatus &= ATAERR_HWERROR;
				pATATask->ATAStatus |= ATA_ERROR;						
	
				/* read 3f6 */
				for ( i = 0; i < 100; i++ ) {
					if (!(ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEAlternateStatus) & IDE_STATUS_BUSY))
						break; /* wait a while for ATAPI device */
				}
				GetStatus(ATABaseAddr[pATATask->bCh2Chip], pATATask->Cmd.ATACmd.bStatus);
								
				if (pATATask->Cmd.ATACmd.bStatus & IDE_STATUS_BUSY){
					pATATask->ATAStatus &= ~ATAERR_DRV;
					pATATask->ATAStatus |= ATA_TIMEOUT;
				}
				else{				
					/* set disk status */
					if (pATATask->ATACmdFlag & ATA_DEVICE){									
						ATAErrorStatus(pATATask);
						
						pATATask->Cmd.ATACmd.bStatus = 0x0; 
						pATATask->Cmd.ATACmd.bError = 0x0;
		
						/* _ERR_INJECT_ */
						switch(ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.ErrType){
							case ERRINJ_ATA_IDNF:
								pATATask->Cmd.ATACmd.bStatus = 0x51; 
								pATATask->Cmd.ATACmd.bError = 0x14;
								break;
							case ERRINJ_ATA_UNC:
								pATATask->Cmd.ATACmd.bStatus = 0x51; 
								pATATask->Cmd.ATACmd.bError = 0x44;
								pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
								break;	
							case ERRINJ_ATA_UNC_W:
								if(pATATask->ATACmdFlag & camDATA_IN){
									pATATask->Cmd.ATACmd.bStatus = 0x51; 
									pATATask->Cmd.ATACmd.bError = 0x44;
									pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
								}
								else{
									pATATask->Cmd.ATACmd.bStatus = 0x70; 
									pATATask->Cmd.ATACmd.LBAl = ErrInj_Buffer[ErrInj_TaskId].ErrInjTask_Config.HitLBAl;
								}
								break;
							case ERRINJ_ATA_CRC:
								pATATask->Cmd.ATACmd.bStatus = 0x51; 
								pATATask->Cmd.ATACmd.bError = 0x84;
								break;
							case ERRINJ_ATA_DF:
								pATATask->Cmd.ATACmd.bStatus = 0x70; 
								break;
							default:	
								break;
						}									
					}
					else if (pATATask->ATACmdFlag & SWAPBOX_CMD)
						ATAErrorStatus(pATATask);
					else { 	/* ATAPI device */
						if ( (!(pATATask->ATAStatus &ATAERR_DRV)) && ( pATATask->ATAStatus & ATAERR_OVERRUN ) )
						{	/* that is not a device error */
							pATATask->ATAStatus = 0;
							goto SuccessCommandCompleted;
						}
						if (!(pATATask->ATACmdFlag & INTERNAL_SENSE)){
							/* get ATAPI error , if INTERNAL_SENSE error, ingore it.*/
							ATAPIErrorStatus(pATATask);
								
							/* must have sense buffer so can do internel sense command */
							if ((pATATask->SenseInfoLength > 1)){
								/* If ATAPI packet command error, than send a internel "REQUEST SENSE Command" 
								     to get the sense datas. 	*/
								pATATask->ATACmdFlag = INTERNAL_SENSE;
								IntReason = INT_WAITCMD;
								goto SuccessCommandCompleted;
							}
						}
						else {
							pATATask->ATACmdFlag &= ~INTERNAL_SENSE;
							ATAPIErrorStatus(pATATask); /* even request sense is failed. */
							ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEStatus); /*clear possible device INT */
						}											
					}
				}
				goto FailCommandCompleted;
			}
			else
			{
				pCurSEQ_INFO->PACCNT--;
			}
		}
		else { /* success done all seq's packets */
			pATATask->ATAStatus= 0;
			
			/* get INT reason */
			if (!(pATATask->ATACmdFlag & ATA_DEVICE))
				IntReason = ReadIndexUchar(ATABaseAddr[bChannelNumber], iNDEXIDESectorCount)& 0x07;
				
			/* If the packet's ATACmdFlag be set with ATA_STATUS_RETURN , it only can queued at the tail of execute packets queue(or only itself at the queue),
    				   so when error or success completed, will return ata registers statses.*/
			if (pATATask->ATACmdFlag & ATA_STATUS_RETURN) /* need read it's ata registers statses when command success completed */
				ATAErrorStatus(pATATask);
		}
SuccessCommandCompleted:					
				
					
		if(IntReason == INT_WAITCMD){
			FreePacketSource = FALSE;
			SubmitInternelSense(pATATask,ultmpATAPacketAddress);
		}
		else if ( pATATask->ATACmdFlag & PIO_XFER )
		{
			if (!(pATATask->ATACmdFlag & ATA_DEVICE) ) { /* ATAPI */
				U32	DataTransferLengthTotal=0;
				U32	DataTransferLengthTotalTemp=0;
				U8 PIOStatus=camSUCCESS;
				
				bdata = ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDEStatus); /* have data to be trasfer */ /*clear possible device INT */
				if (bdata & IDE_STATUS_DRQ){
					
					if ( gChnlConfig[bChannelNumber].ChnlType == CHNL_SATA ) {
						ReadIndexUlong(ATA_ExpBaseAddr[bChannelNumber], iNDEXeXP_PIO_Transfer_count, DataTransferLengthTotalTemp);
						DataTransferLengthTotalTemp &= 0x0000FFFF;									
					}
					else {
						DataTransferLengthTotalTemp = (U32)ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECylinderLow)
				                   | ((U32)ReadIndexUchar(ATABaseAddr[pATATask->bCh2Chip], iNDEXIDECylinderHigh) << 8);
					}
					DataTransferLengthTotalTemp >>= 1;  
					DataTransferLengthTotal = DataTransferLengthTotal+DataTransferLengthTotalTemp;

					if (DataTransferLengthTotalTemp==0) { /* no datas to be transfer */
						/* soft reset ATA Module after getting status. */
						ATASoftResetModule(bChannelNumber);
						pATATask->ATAStatus = ATA_ERROR;
						errorFlag=TRUE;
						((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;
						ATAModuleSEQID[bChannelNumber] = U8NULL;									
						goto FailCommandCompleted;
					}

					if ( pATATask->ATACmdFlag & camDATA_IN ){
						
						PIOStatus = ReadBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLengthTotalTemp );
					}
					else {
						PIOStatus = WriteBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLengthTotalTemp );
						}
					
					/* check error from Sequence Counter Control Register */
					ReadPortUchar(HostBaseAddr[Adapter_ID], (SEQID*4)+2, SEQCntCtrlErr);
					SEQCntCtrlErr &= (1<<ModuleType);
					
					/* after PIO data transfer, maybe be with error */
					if (SEQCntCtrlErr || (PIOStatus == camFAIL)) {
						/* soft reset ATA Module after getting status. */
						ATASoftResetModule(bChannelNumber);
						pATATask->ATAStatus = ATA_ERROR;
						errorFlag=TRUE;
						((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;
						ATAModuleSEQID[bChannelNumber] = U8NULL;									
						goto FailCommandCompleted;
					}
										
					(PU16)pATATask->DataBuffer = (PU16)pATATask->DataBuffer+DataTransferLengthTotal;

					/* need next ISR to clear ATAPI interrupt */
					pCurSEQ_INFO->SEQID2PacketHead=(PVOID)ultmpATAPacketAddress;
					pCurSEQ_INFO->SEQID2PacketTail=(PVOID)ultmpATAPacketAddress;									
					return;
					/*((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;*/
					/*ATAModuleSEQID[bChannelNumber] = U8NULL;*/
				}
				else{/* else of if (bdata & IDE_STATUS_DRQ) */
					if ( pATATask->ATACmdFlag & INTERNAL_SENSE )
						pATATask->ATAStatus = ATA_ERROR + ATAERR_DRV;
					pATATask->ATACmdFlag &= ~INTERNAL_SENSE;
					((PPACKET_HEAD)ultmpATAPacketAddress)->Flag &= ~PACFLAG_ATAPI_PIO;
					ATAModuleSEQID[bChannelNumber] = U8NULL;	
				}
			}
			else { /* ATA */
				U32	DataTransferLength = 0;
				U8 PIOStatus;
				
				DataTransferLength = (pATATask->DataTransferLength >> 1);
				if ( pATATask->ATACmdFlag & camDATA_IN ){
					PIOStatus = ReadBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLength );
				}
				else {
					PIOStatus = WriteBuffer(pATATask->bCh2Chip,(PU16)pATATask->DataBuffer, DataTransferLength );
				}		
			
				if(PIOStatus == camFAIL)
					pATATask->ATAStatus = ATA_ERROR;
			}
		}	/*end of (pATATask->ATACmdFlag & PIO_XFER )*/									

FailCommandCompleted:				
		if (FreePacketSource==TRUE) {
			/* free sg resource */
			CAMFreeATASG(bChannelNumber, (PVOID)ultmpATAPacketAddress->PacketHead.PSG);	
				/* free packet resource. */
			CAMFreeCamPacket(bChannelNumber, (PVOID)ultmpATAPacketAddress);
		}
								
		if ( !(pATATask->ATACmdFlag & INTERNAL_SENSE) ) {
			pATATask->callback(pATATask);	
		}	

		ultmpATAPacketAddress_next = (PCAM_PACKET)ultmpATAPacketAddress->PacketHead.NPA;
		
		if (errorFlag==TRUE)
			break;
			
		/* next packet */
		ultmpATAPacketAddress = ultmpATAPacketAddress_next;
		
	}/* end of while ((U32)ultmpATAPacketAddress != PTR_NULL)*/

	if (RestartModule==FALSE) 
		ATAModuleBusy[bChannelNumber]=FALSE;

	CAMFreeSEQID(Adapter_ID,SEQID);	

	/* must check module is need to push, By it maybe pushed when no SQEID. */
	for (i=(Adapter_ID*MAX_ATA_CHNL);i<((Adapter_ID+1)*MAX_ATA_CHNL);i++) {
		if ( (ATAModuleNeedPush[i]==TRUE) && (ATAModuleBusy[i]==FALSE) ){
			/* start ata module */
			StartATAOperation(i);
			break;
		}
	}



}




#endif

